"""
Here is the common models for all the applications.

This will be shared through all the dockers and can be accesses by importing it.
"""
from django.db import models


class ChargerTypeM2M(models.Model):
    location_charger = models.ForeignKey(
        'LocationCharger', on_delete=models.CASCADE)
    charger_location_type = models.ForeignKey(
        'ChargerLocationType', on_delete=models.CASCADE)


class ChargerVelocityM2M(models.Model):
    location_charger = models.ForeignKey(
        'LocationCharger', on_delete=models.CASCADE)
    charger_velocity = models.ForeignKey(
        'ChargerVelocity', on_delete=models.CASCADE)


class LocationCharger(models.Model):
    """
    Model for storing the location of the chargers.
    """

    # General info
    promotorGestor = models.CharField(max_length=100)
    access = models.CharField(max_length=100)

    # Charger type info
    connectionType = models.ManyToManyField(
        'ChargerLocationType', related_name='connectionType', through="ChargerTypeM2M")
    kw = models.FloatField()
    acDc = models.CharField(max_length=5)
    velocities = models.ManyToManyField(
        'ChargerVelocity', related_name='velocities', through="ChargerVelocityM2M")

    # Location
    latitud = models.FloatField()
    longitud = models.FloatField()
    adreA = models.CharField(max_length=100)

    class Meta:
        app_label = "common"


class ChargerVelocity(models.Model):

    NORMAL = 'NORMAL'
    semiRAPID = 'semiRAPID'
    RAPID = 'RAPID'
    superRAPID = 'superRAPID'

    VELOCITY_CHOICES = [
        (NORMAL, 'NORMAL'),
        (semiRAPID, 'semiRAPID'),
        (RAPID, 'RAPID'),
        (superRAPID, 'superRAPID')
    ]
    velocity = models.CharField(
        max_length=10, choices=VELOCITY_CHOICES, unique=True)

    class Meta:
        app_label = "common"

    def str(self):
        return self.velocity


class ChargerLocationType(models.Model):
    """
    Model to represent the types of chargers.
    """

    MENNEKES = "MENNEKES"
    TESLA = "TESLA"
    SCHUKO = "SCHUKO"
    CHADEMO = "CHADEMO"
    CCS_COMBO2 = "CCS COMBO2"

    CHARGER_CHOICES = [
        (MENNEKES, "MENNEKES"),
        (TESLA, "TESLA"),
        (SCHUKO, "SCHUKO"),
        (CHADEMO, "CHADEMO"),
        (CCS_COMBO2, "CCS COMBO2"),
    ]

    chargerType = models.CharField(
        max_length=20, choices=CHARGER_CHOICES, unique=True)

    def __str__(self):
        return self.chargerType

    class Meta:
        app_label = "common"



"""
Here is the common models for all the applications.

This will be shared through all the dockers and can be accesses by importing it.

__example: from common.models import User, Driver
"""

from django.contrib.auth.models import User as baseUser
from django.core.validators import MinValueValidator
from django.db import models
from storages.backends.s3boto3 import S3Boto3Storage


class User(baseUser):
    """
    User class for our project

    Args:
        baseUser (User): default django implementation for user
            (see the documentation for more info)
    """

    Google = "google"
    Facebook = "facebook"
    Base = "base"

    typeLoginChoices = [(Google, "google"), (Facebook, "facebook"), (Base, "base")]

    # change to keep the pk defined in the UML consistent
    baseUser.email = models.EmailField("email address", unique=True)  # type: ignore
    birthDate = models.DateField(auto_now=False, auto_now_add=False)
    points = models.IntegerField(default=0)
    updatedAt = models.DateTimeField(
        "Last modification of the User", auto_now=True, auto_now_add=False
    )
    createdAt = models.DateTimeField("Creation date of the User", auto_now=False, auto_now_add=True)

    profileImage = models.ImageField(
        upload_to="profile_image/",
        null=True,
        blank=True,
        default="default.png",
        storage=S3Boto3Storage(location="media/profile_image"),
    )

    typeOfLogin = models.CharField(max_length=50, choices=typeLoginChoices, default=Base)

    class Meta:
        app_label = "common"


class ChargerType(models.Model):
    """
    Model to represent the types of chargers.
    """

    MENNEKES = "MENNEKES"
    TESLA = "TESLA"
    SCHUKO = "SCHUKO"
    CHADEMO = "CHADEMO"
    CCS_COMBO2 = "CCS COMBO2"

    CHARGER_CHOICES = [
        (MENNEKES, "MENNEKES"),
        (TESLA, "TESLA"),
        (SCHUKO, "SCHUKO"),
        (CHADEMO, "CHADEMO"),
        (CCS_COMBO2, "CCS COMBO2"),
    ]

    chargerType = models.CharField(max_length=20, choices=CHARGER_CHOICES, unique=True)

    def __str__(self):
        return self.chargerType

    class Meta:
        app_label = "common"


class Preference(models.Model):
    """
    Model to represent driver preferences.
    """

    canNotTravelWithPets = models.BooleanField(default=False)
    listenToMusic = models.BooleanField(default=False)
    noSmoking = models.BooleanField(default=False)
    talkTooMuch = models.BooleanField(default=False)

    def __str__(self):
        return (
            "Travel Preferences: Can't Travel With Pets - "
            + str(self.canNotTravelWithPets)
            + ", Listen to Music - "
            + str(self.listenToMusic)
            + ", No Smoking - "
            + str(self.noSmoking)
            + ", Talk Too Much - "
            + str(self.talkTooMuch)
        )

    class Meta:
        app_label = "common"


class Driver(User):
    """
    This is the driver class
    """

    # Pointer to the parent class witch we need since if not done
    # it will end up referencing to baseUser and it will give conflicts.
    # Also note that this will use the default UserManager thus the
    # calls in the serializer create function and the default ones, will reference the
    # UserManager functions such as Driver.objects.create_user instead of
    # Driver.objects.create_driver.
    # If more precise creation is needed the DriverManager class should be implemented
    parent_pointer = models.OneToOneField(
        "User", verbose_name="Parent Class", on_delete=models.CASCADE, parent_link=True
    )

    # Rest of the fields needed
    dni = models.CharField(max_length=50, unique=True)
    driverPoints = models.IntegerField(default=0)
    autonomy = models.IntegerField(default=1, validators=[MinValueValidator(1)])

    # Charger type attributes
    chargerTypes = models.ManyToManyField("ChargerType", related_name="drivers")

    # Preferences attributes
    preference = models.OneToOneField("Preference", on_delete=models.CASCADE, null=True)

    iban = models.CharField(max_length=36, unique=True, blank=True)

    def save(self, *args, **kwargs) -> None:
        """
        Override of the save method to add the preference if it does not exist
        """
        if not hasattr(self, "preference") or not self.preference:
            self.preference = Preference.objects.create()

        return super().save(*args, **kwargs)

    class Meta:
        app_label = "common"


class Report(models.Model):
    """
    Model for storing Reports given by users to users or drivers.
    """

    reporter = models.ForeignKey(User, related_name="report_giver", on_delete=models.CASCADE)
    reported = models.ForeignKey(User, related_name="report_receiver", on_delete=models.CASCADE)

    updatedAt = models.DateTimeField(
        "Last modification of the Report", auto_now=True, auto_now_add=False
    )
    createdAt = models.DateTimeField(
        "Creation date of the Report", auto_now=False, auto_now_add=True
    )

    comment = models.TextField(blank=True)

    def __str__(self):
        return f"{self.reporter} -> {self.reported}"

    class Meta:
        app_label = "common"



"""
Data models definition for all PowerPathFinder services, this module serves as single source of
truth. Other services will introspect the DB in order to be up to date. It includes the following
models:

- Route: Describes the routes that drivers can create and passengers can join.
- RoutePassengers: Describes the passengers that are part of a route, records are created when
    users join to a route.
"""

from datetime import timedelta
from django.db import models
from django.db.models import Q

from .user import Driver, User


class RouteManager(models.Manager):
    """
    RouteManager is a custom Manager that adds 'table-level' functionality
    This managers will act as the entry point of the domain to the persistance layer
    https://docs.djangoproject.com/en/5.0/topics/db/managers/#custom-managers
    # See QuerySet API https://docs.djangoproject.com/en/5.0/ref/models/querysets/#queryset-api
    # See Filtered Relations https://docs.djangoproject.com/en/5.0/ref/models/querysets/#queryset-api
    # See Complex Lookups https://docs.djangoproject.com/en/5.0/topics/db/queries/#complex-lookups-with-q-objects
    """

    def byUser(self, userId):
        """
        returns all routes where the user is present either as driver or passenger
        """
        return self.get(Q(driver_id=userId) | Q(passengers__id__contains=userId))

    def active(self):
        """
        active queryset, retrieves all routes that ar not either cancelled or finalized
        """
        return self.filter(cancelled=False, finalized=False)

    def cancelled(self):
        """
        cancelled queryset, retrieves all routes that are cancelled
        """
        return self.filter(cancelled=True)

    def finalized(self):
        """
        finalized queryset, retrieves all routes that are finalized
        """
        return self.filter(finalized=True)

    def withSeats(self):
        """
        withSeats queryset, retrieves all routes that has at least one free seat
        """
        return self.filter(freeSeats__gt=0)


class Route(models.Model):
    """
    Route between two points, organized by a driver to be shared with passengers.
    """

    objects: RouteManager = RouteManager()  # Must assign the custom manager and specify the type

    driver = models.ForeignKey(Driver, on_delete=models.CASCADE)

    originLat = models.FloatField()
    originLon = models.FloatField()
    originAlias = models.CharField(max_length=100)

    destinationLat = models.FloatField()
    destinationLon = models.FloatField()
    destinationAlias = models.CharField(max_length=100)

    polyline = models.TextField()
    waypoints = models.JSONField(default=list)
    distance = models.PositiveIntegerField()
    duration = models.PositiveIntegerField()

    departureTime = models.DateTimeField()
    freeSeats = models.PositiveSmallIntegerField()
    price = models.FloatField(default=0.0)

    passengers = models.ManyToManyField(User, related_name="joined_routes")

    cancelled = models.BooleanField(default=False)
    finalized = models.BooleanField(default=False)

    createdAt = models.DateTimeField(auto_now_add=True)

    class Meta:
        app_label = "common"

    def isFull(self):
        """
        Returns True if the route is full, False otherwise.
        """
        return self.freeSeats == 0

    def overlapsWith(self, routeId):  # TODO refactor to accept Route instance
        """
        Returns True if the route temporally overlaps with the route with the provided ID, False otherwise.
        """
        duration = timedelta(seconds=self.duration)
        route = Route.objects.get(id=routeId)
        if self.departureTime + duration >= route.departureTime:
            return True
        return False



"""
Here is the common models for all the applications.

This will be shared through all the dockers and can be accesses by importing it.

__example: from common.models import Valuation
"""

from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from .user import User
from .route import Route


class Valuation(models.Model):
    """
    Model for storing valuations given by users to users or drivers.
    """

    RATING_CHOICES = [
        (1, "1"),
        (2, "2"),
        (3, "3"),
        (4, "4"),
        (5, "5"),
    ]

    giver = models.ForeignKey(User, related_name="given_valuations", on_delete=models.CASCADE)
    receiver = models.ForeignKey(
        User,
        related_name="received_user_valuations",
        on_delete=models.CASCADE,
    )
    route = models.ForeignKey(Route, related_name="route_valuations", on_delete=models.CASCADE)
    rating = models.IntegerField(
        choices=RATING_CHOICES, validators=[MinValueValidator(1), MaxValueValidator(5)]
    )
    comment = models.TextField(blank=True)
    createdAt = models.DateTimeField(auto_now_add=True)

    class Meta:
        """
        Meta used to add the label so that the imports work correctly
        """

        app_label = "common"
        unique_together = ("giver", "receiver", "route")



"""
Here is the common models for all the applications.

This will be shared through all the dockers and can be accesses by importing it.

__example: from common.models import Payment
Payment: Maintains a record of the payments made by users for routes.
"""

from django.db import models

from .user import User
from .route import Route


class Payment(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    route = models.ForeignKey(Route, on_delete=models.CASCADE)
    amount = models.DecimalField(max_digits=10, decimal_places=2)
    date = models.DateTimeField(auto_now_add=True)
    description = models.CharField(max_length=100)
    paymentIntentId = models.CharField(max_length=100, unique=True)
    isRefunded = models.BooleanField(default=False)

    def __str__(self):
        return f"{self.user.username}'s payment of {self.amount} on {self.date}"

    class Meta:
        """
        Meta used to add the label so that the imports work correctly
        """

        app_label = "common"



"""
Here is the common models for all the applications.

This will be shared through all the dockers and can be accesses by importing it.

__example: from common.models.achievement import Achievement
"""

from django.db import models
from .user import User


class Achievement(models.Model):
    title = models.CharField(max_length=100, unique=True)
    description = models.TextField()
    required_points = models.IntegerField()

    class Meta:
        app_label = "common"


class UserAchievementProgress(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    achievement = models.ForeignKey(Achievement, on_delete=models.CASCADE)
    progress = models.IntegerField(default=0)
    achieved = models.BooleanField(default=False)
    date_achieved = models.DateTimeField(null=True, blank=True)

    class Meta:
        app_label = "common"
        unique_together = ("user", "achievement")



from django.db import models


class FCMToken(models.Model):
    """
    Model to represent the FCM token.
    """

    user = models.OneToOneField("User", on_delete=models.CASCADE, related_name="fcm_token")
    token = models.CharField(unique=True, max_length=255, null=True)
    ts = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return self.token

    class Meta:
        app_label = "common"
